# Client errors
BAD_REQUEST = 400
NOT_FOUND = 404
LENGTH_REQUIRED = 411
UNSUPPORTED_MEDIA_TYPE = 415
TYPE_ERROR = 452
INVALID_ARGUMENT = 453


# Server errors
SERVER_ERROR = 500
SERVICE_UNAVAILABLE = 503
